
#include "data_safe.h"

/* Data Layout:

   keys[] stores the keys that are currently in use. Since -1 is not a
   legal key, we use it to mark empty places here, i.e. if
   keys[key]=-1, we can store a new key there.

   key_entries[] stores the number of elements stored under a key,
   i.e. key_entries[key] is the the number of values stored for
   keys[key]. 

   This would be better implemented as one array of  type 
   struct{int keys; int entries;}, but structs were not known when the
   assignment was given.  
   
   key_array[], data_array[] and entry_no_array[] together represent
   the real data safe. Again, this would be nicer as a single array of
   struct {int key, int value, int entrie_ne}

   The first is the key, the second is the value, the third is the
   running number of the entry for that key, i.e. 1 for the first call
   of data_safe(ds_store, key, val), 2 for the second, ... */


int data_safe(DS_Operation method, int key, int value_or_index)
{
   static int initialized = 0;
   static int keys[CONCURRENT_KEYS]; /* Keys that are in use */
   static int keys_in_use = 0;
   static int key_entries[CONCURRENT_KEYS]; /* How many values are
					     * stored under that key?
					     * Redundant, but makes
					     * coding easier */
   static int key_array[DATA_SAFE_SIZE];  /* Whose data is this? */
   static int total_entries = 0;
   static int data_array[DATA_SAFE_SIZE]; /* What is the data? */
   static int entry_no_array[DATA_SAFE_SIZE]; /* What's the running
					       * number of that entry
					       * for the key? */
   int i, key_index;

   if(!initialized)
   {
      for(i=0; i<CONCURRENT_KEYS; i++)
      {
	 keys[i] = -1; /* Unused */	 
      }
      for(i=0; i<DATA_SAFE_SIZE; i++)
      {
	 key_array[i] = -1; /* Unused */	 
      }
      initialized = 1;
      /* Note: There is no need to initialize the rest, as all this
	 data will be filled in by any relevant operation anyways */
   }
   switch(method)
   {
   case ds_register:
	 if(keys_in_use == CONCURRENT_KEYS)
	 {
	    printf("Out of key entries!\n");
	    return -1;
	 }
	 key = -1;
	 while(key == -1)
	 {
	    key = rand()%12344321; /* Make it fit */
	    /* printf("Trying key %d\n", key); */
	    for(i=0; i<CONCURRENT_KEYS && key!=-1; i++)
	    {
	       if(keys[i] == key)
	       { /* Key is in use, try again */
		  key = -1;
	       }
	    }
	 }
	 /* Note: Here, key is a key not in key_array */
	 
	 key_index = -1;
	 for(i=0; i<CONCURRENT_KEYS && key_index == -1; i++)
	 {
	    if(keys[i] == -1)
	    {
	       key_index = i; /* Found an empty spot! */
	    }
	 }
	 /* key_index points to a position in key_array that is empty */
	 /* printf("Key %d stored at %d\n", key, key_index); */
	 keys[key_index] = key;
	 key_entries[key_index] = 0;
	 keys_in_use++;
	 return key;
	 break; /* Redundant, but good practice */
   case ds_store:
	 if(total_entries == DATA_SAFE_SIZE)
	 { /* No space left */
	    return -1;
	 }
	 key_index = -1;
	 for(i=0; i<CONCURRENT_KEYS && key_index == -1; i++)
	 {
	    if(keys[i] == key)
	    { 
	       key_index = i; /* Key is valid! */
	    }
	 }
	 if(key_index == -1)
	 { /* Key is not valid */
	    return -1;
	 }
	 /* printf("Key %d is valid\n", key); */
	 for(i=0; i<DATA_SAFE_SIZE; i++) /* Find empty spot */
	 { 
	    if(key_array[i] ==-1) /* Empty spot? */
	    {
	       total_entries++;
	       key_entries[key_index]++;
	       key_array[i] = key;
	       data_array[i] = value_or_index;
	       entry_no_array[i] = key_entries[key_index];
	       /* printf("Stored %d as entry %d for key%d\n", 
		  value_or_index, key, key_entries[key_index]); */
	       return value_or_index;
	    }
	 }
	 break;
   case ds_retrieve:
	 for(i=0; i<DATA_SAFE_SIZE; i++) /* Look for entry */
	 { 
	    if(key_array[i] == key && key_entries[i] == value_or_index)
	    {
	       /* Found it! */
	       return data_array[i];
	    }
	 }
	 return -1; /* Not found */;
	 break;
   case ds_delete:
	 for(i=0; i<CONCURRENT_KEYS; i++)
	 {
	    if(keys[i] == key)
	    {
	       keys[i] = -1;
	    }
	 }
	 for(i=0; i<DATA_SAFE_SIZE; i++)
	 {
	    if(key_array[i] == key)
	    {
	       key_array[i] = -1;
	    }
	 }
	 keys_in_use--;
	 return key;
   }   
   return -1;
}
